/*
	This file is part of lgui.

	lgui is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	lgui is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with lgui.  If not, see <http://www.gnu.org/licenses/>.

    Copyright (C) 2008 Lucas Hermann Negri
*/

#include "proxylist.h"

static void lgui_proxy_list_init(LguiProxyList* model);

static void lgui_proxy_list_class_init(LguiProxyListClass* klass);

static void lgui_proxy_list_tree_model_init(GtkTreeModelIface* iface);

static void lgui_proxy_list_finalize(GObject* object);

static GtkTreeModelFlags lgui_proxy_list_get_flags(GtkTreeModel* model);

static gint lgui_proxy_list_get_n_columns(GtkTreeModel* model);

static GType lgui_proxy_list_get_column_type(GtkTreeModel* model, gint index);

static gboolean lgui_proxy_list_get_iter(GtkTreeModel* model, GtkTreeIter *iter,
	GtkTreePath* path);
	
static GtkTreePath* lgui_proxy_list_get_path(GtkTreeModel* model, GtkTreeIter* iter);

static void lgui_proxy_list_get_value(GtkTreeModel* model, GtkTreeIter* iter,
	gint column, GValue* value);
	
static gboolean lgui_proxy_list_iter_next(GtkTreeModel* model, GtkTreeIter* iter);

static gboolean lgui_proxy_list_iter_children(GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* parent);
	
static gboolean lgui_proxy_list_iter_has_child(GtkTreeModel* model,	GtkTreeIter* iter);

static gint lgui_proxy_list_iter_n_children(GtkTreeModel* model, GtkTreeIter* iter);

static gboolean lgui_proxy_list_iter_nth_child(GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* parent, gint n);
	
static gboolean lgui_proxy_list_iter_parent(GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* child);

static GObjectClass* parent_class = NULL;

GType lgui_proxy_list_get_type()
{
	static GType lgui_proxy_list_type = 0;

	if(lgui_proxy_list_type == 0)
	{
    	static const GTypeInfo lgui_proxy_list_info =
		{
			sizeof(LguiProxyListClass),
			NULL,									/* base_init */
			NULL,									/* base_finalize */
			(GClassInitFunc)lgui_proxy_list_class_init,
			NULL,									/* class finalize */
			NULL,									/* class_data */
			sizeof(LguiProxyList),
			0,										/* n_preallocs */
			(GInstanceInitFunc)lgui_proxy_list_init
    	};
	
		static const GInterfaceInfo tree_model_info =
		{
			(GInterfaceInitFunc)lgui_proxy_list_tree_model_init,
			NULL,
			NULL
		};

		lgui_proxy_list_type = g_type_register_static(G_TYPE_OBJECT, "LguiProxyList",
			&lgui_proxy_list_info, (GTypeFlags)0);
		
		g_type_add_interface_static(lgui_proxy_list_type, GTK_TYPE_TREE_MODEL, &tree_model_info);
	}

	return lgui_proxy_list_type;
}

static void lgui_proxy_list_class_init(LguiProxyListClass* klass)
{
	GObjectClass* object_class;

	parent_class = (GObjectClass*)g_type_class_peek_parent(klass);
	object_class = (GObjectClass*)klass;

	object_class->finalize = lgui_proxy_list_finalize;
}

static void lgui_proxy_list_tree_model_init(GtkTreeModelIface* iface)
{
	iface->get_flags       = lgui_proxy_list_get_flags;
	iface->get_n_columns   = lgui_proxy_list_get_n_columns;
	iface->get_column_type = lgui_proxy_list_get_column_type;
	iface->get_iter        = lgui_proxy_list_get_iter;
	iface->get_path        = lgui_proxy_list_get_path;
	iface->get_value       = lgui_proxy_list_get_value;
	iface->iter_next       = lgui_proxy_list_iter_next;
	iface->iter_children   = lgui_proxy_list_iter_children;
	iface->iter_has_child  = lgui_proxy_list_iter_has_child;
	iface->iter_n_children = lgui_proxy_list_iter_n_children;
	iface->iter_nth_child  = lgui_proxy_list_iter_nth_child;
	iface->iter_parent     = lgui_proxy_list_iter_parent;
}

static void lgui_proxy_list_init(LguiProxyList* proxy_list)
{
	proxy_list->L = NULL;
	proxy_list->table = LUA_REFNIL;
	proxy_list->stamp = g_random_int();
}

static void lgui_proxy_list_finalize(GObject *object)
{
	LguiProxyList* list = LGUI_PROXY_LIST(object); 
	luaL_unref(list->L, LUA_REGISTRYINDEX, list->table);
  	list->table = LUA_REFNIL;
	g_free(list->column_types);

	(*parent_class->finalize)(object);
}

static GtkTreeModelFlags lgui_proxy_list_get_flags(GtkTreeModel* model)
{
	return (GTK_TREE_MODEL_LIST_ONLY | GTK_TREE_MODEL_ITERS_PERSIST);
}

void lgui_proxy_list_get_method(LguiProxyList* list, const char* name)
{
	#ifdef IDEBUG
	g_debug("Getting method %s!", name);
	#endif
	
	/* Get the function */
	lua_rawgeti(list->L, LUA_REGISTRYINDEX, list->table);
    lua_getfield(list->L, -1, name);
	
	if(!lua_isfunction(list->L, -1))
		luaL_error(list->L, "The method %s is required", name);
	
	/* Push the "self" */
	lua_pushvalue(list->L, -2);
}

static gint lgui_proxy_list_get_n_columns(GtkTreeModel* model)
{
	LguiProxyList* list = LGUI_PROXY_LIST(model);
	return list->n_columns;
}

static GType lgui_proxy_list_get_column_type(GtkTreeModel* model, gint index)
{
	return G_TYPE_STRING;
}

static gboolean lgui_proxy_list_get_iter(GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreePath* path)
{
	LguiProxyList* proxy_list = LGUI_PROXY_LIST(model);
	
	gint* indices = gtk_tree_path_get_indices(path);
	iter->stamp = proxy_list->stamp;
	iter->user_data = GINT_TO_POINTER(indices[0]);
	iter->user_data2 = iter->user_data3 = NULL;
	
	return TRUE;
}

static GtkTreePath* lgui_proxy_list_get_path(GtkTreeModel* model, GtkTreeIter* iter)
{
	GtkTreePath* path = gtk_tree_path_new();
    gtk_tree_path_append_index(path, GPOINTER_TO_INT(iter->user_data));
		
	return path;
}

static void lgui_proxy_list_get_value(GtkTreeModel* model, GtkTreeIter* iter,
	gint column, GValue* value)
{
	LguiProxyList* proxy_list = LGUI_PROXY_LIST(model);
	lua_State *L = proxy_list->L;

	int top = lua_gettop(L);
    lgui_proxy_list_get_method(proxy_list, "getValue");
	lua_pushnumber(L, GPOINTER_TO_INT(iter->user_data));
    lua_pushnumber(L, column);
	lua_call(L, 3, 1);
	 
    /* Fill the GValue */
	lua_getfield(L, LUA_GLOBALSINDEX, "__lobjValueSet");
	g_value_init(value, proxy_list->column_types[column]);
	lua_pushlightuserdata(L, value);
	lua_pushvalue(L, -3); /* push the returned value from getValue */
	lua_call(L, 2, 0);
	
    lua_settop(L, top);
}

static gboolean lgui_proxy_list_iter_next (GtkTreeModel* model, GtkTreeIter* iter)
{
	LguiProxyList* proxy_list = LGUI_PROXY_LIST(model);
	lua_State* L = proxy_list->L;
	gboolean ok;
	
	int top = lua_gettop(L);
	
    lgui_proxy_list_get_method(proxy_list, "iterNext");
	gint current = GPOINTER_TO_INT(iter->user_data);
  	lua_pushinteger(L, current);
	lua_call(L, 2, 1);
	 
    if(lua_isboolean(L, -1) && lua_toboolean(L, -1))
	{
		iter->stamp = proxy_list->stamp;
		iter->user_data = GINT_TO_POINTER(current + 1);
		ok = TRUE;
    }
	else
		ok = FALSE;
		
	lua_settop(L, top);
	
	return ok;
}

static gboolean lgui_proxy_list_iter_children (GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* parent)
{
	if(parent)
		return FALSE;
	else
	{
		LguiProxyList* list = LGUI_PROXY_LIST(model);
		iter->stamp = list->stamp;
		iter->user_data = GINT_TO_POINTER(1);
		return TRUE;
	}
}

static gboolean lgui_proxy_list_iter_has_child (GtkTreeModel* model, GtkTreeIter* iter)
{
	return FALSE;
}

static gint lgui_proxy_list_iter_n_children(GtkTreeModel* model, GtkTreeIter* iter)
{
	if(!iter)
    	return lgui_proxy_list_get_n_columns(model);
	
	return 0;
}

static gboolean lgui_proxy_list_iter_nth_child (GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* parent, gint n)
{
	if(parent)
		return FALSE;
	else
	{
		LguiProxyList* proxy_list = LGUI_PROXY_LIST(model);
		iter->stamp = proxy_list->stamp;
		iter->user_data = GINT_TO_POINTER(n);
		return TRUE;
	}
}

static gboolean lgui_proxy_list_iter_parent (GtkTreeModel* model, GtkTreeIter* iter,
	GtkTreeIter* child)
{
	return FALSE;
}

int lgui_proxy_list_new(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TTABLE);
	
	int n_columns = lua_gettop(L) - 1, i;
	GType* types = g_malloc(sizeof(GType) * n_columns);

	/* Fill the types array */
	for(i = 1; i <= n_columns; i++)
	{
		types[i - 1] = g_type_from_name(luaL_checkstring(L, i + 1));
	}

	LguiProxyList* ptr = g_object_new(LGUI_TYPE_PROXY_LIST, NULL);
	ptr->L = L;
	ptr->n_columns = n_columns;
	ptr->column_types = types;
	
	lua_pushvalue(L, 1);
	ptr->table = luaL_ref(L, LUA_REGISTRYINDEX);
	
	object_new(L, ptr, TRUE);
	return 1;
}

int lgui_proxy_list_set_value(lua_State* L)
{
	luaL_checktype(L, 1, LUA_TUSERDATA);
	luaL_checktype(L, 2, LUA_TUSERDATA);
	luaL_checktype(L, 3, LUA_TNUMBER);

	Object* obj = lua_touserdata(L, 1);
	Object* stc = lua_touserdata(L, 2);
	
	lgui_proxy_list_get_method(LGUI_PROXY_LIST(obj->pointer), "setValue");
	lua_pushnumber(L, GPOINTER_TO_INT( ((GtkTreeIter*)stc->pointer)->user_data) );
    lua_pushnumber(L, lua_tointeger(L, 3));
    lua_pushvalue(L, 4);
	lua_call(L, 4, 0);
	
	return 0;
}
